#include "pl310.h"

static unsigned int pl310_way_mask;  /* Bitmask of active ways */

static inline void cache_wait_way(unsigned int *reg, unsigned long mask)
{
    /* wait for cache operation by line or way to complete */
    while ( *((volatile unsigned int *)reg) & mask );
}

static inline void cache_sync(void)
{
    *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_CACHE_SYNC)) = 0;
}

static inline void pl310_inv_all(void)
{
    /* invalidate all ways */
    *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_INV_WAY)) = pl310_way_mask;
    cache_wait_way((unsigned int *)(RMA1_L2CACHE_BASE+PL310_INV_WAY), pl310_way_mask);
    cache_sync();
}

void
pl310_init(void)
{
    unsigned int tmp;
    unsigned int aux;
    unsigned int cache_id;
    unsigned int prefetch;
    unsigned int prefetch_val = 0;    
    unsigned int power;
    unsigned int aux_val = 0x40440000;
    unsigned int aux_mask = 0x82000fff;
    int ways;

    /* L2LbVłɃIɂȂĂ邩mF */
    tmp = *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_CTRL));

    /* L2LbV̏ꍇ̂ݏ{ */
    if ( !(tmp & 1) ) {
        cache_id = *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_CACHE_ID));
        aux = *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_AUX_CTRL));
        prefetch = *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_PREFETCH_CTRL));
        power = *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_POWER_CTRL));

        if (aux & (1 << 16))
            ways = 16;
        else
            ways = 8;

        /*
         * Set bit 22 in the auxiliary control register. If this bit
         * is cleared, PL310 treats Normal Shared Non-cacheable
         * accesses as Cacheable no-allocate.
         */
        aux_val |= 1 << 22;
        
        //Inst Prefetch
        aux_val |= 1 << 29;
        prefetch_val |= 1 << 29;

        //Data Prefetch
        aux_val |= 1 << 28;
        prefetch_val |= 1 << 28;

        /*
         * Enable prefetch-related features that can improve system
         * performance.  All bits in the prefetch control register
         * are set to zero by default, and we assume here that no
         * preceding softwares such as bootloaders set up these bits.
         */
        /* safely available in r3p2 or later */
        if ((cache_id & 0x3f) > 0x6)
          prefetch_val |= 1 << 30;
        /* bit 27 and 23 are left unused for now */
                

        pl310_way_mask = (1 << ways) - 1;

        // 8-way 32KB cache, Shared attribute override enable, Early BRESP enable
        aux &= aux_mask;
        aux |= aux_val;
        prefetch |= prefetch_val;
        
        /* pl310 controller is disabled */
        *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_AUX_CTRL)) = aux;
        *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_PREFETCH_CTRL)) = prefetch;
        *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_POWER_CTRL)) = power;
        
        pl310_inv_all();
        
        /* enable PL310 */
        *((volatile unsigned int *)(RMA1_L2CACHE_BASE + PL310_CTRL)) = 1;
    }
    
}
